import sys
import numpy as np
import matplotlib.pyplot as plt
import pickle
import os
import gzip

from torch.utils import data
import torch
import torch.optim as optim
from torch.autograd import Variable
from time import gmtime, strftime
import sys
import torch.nn as nn
from torch.utils.data import DataLoader
from PIL import Image
from data_utils import DatasetFromFolder_mrc_test
from data_utils import save_image, save_mrc

from torchvision import transforms
import math


import model
import globalvar as gl
import config
import os
import mrcfile
from skimage import measure, color,segmentation
import time
from sklearn.cluster import MeanShift, estimate_bandwidth
import cv2


import torchio as tio
from torchio.transforms import (
    RescaleIntensity,
    RandomFlip,
    Compose,
)

class_number = gl.get_value('class_number')
num_layers = gl.get_value('num_layers')
kernel_size = gl.get_value('kernel_size')
batch_size = gl.get_value('batch_size')
learning_rate = gl.get_value('learning_rate')
grow_rate = gl.get_value('grow_rate')
dilation_mod =  gl.get_value('dilation_mod')
log_dir = gl.get_value('log_dir')
modelsave_dir = gl.get_value('modelsave_dir')
epochs = gl.get_value('epochs')
device_ids =[0] #gl.get_value('device_ids')


def global_accuracy(predicted_img, labels):
    total = labels.shape[0] * labels.shape[1] * labels.shape[2]#* labels.shape[3]
    correct = (predicted_img == labels).sum().item()
    return correct/total

def mean_accuracy(predicted_img, labels, class_number):
    correct_list = np.zeros(class_number)
    total_list = np.zeros(class_number)
    for i in range(class_number):
        correct = np.logical_and(labels==predicted_img, labels==i).sum().item()
        correct_list[i] = correct
        total_class = (labels == i).sum().item()
        total_list[i] = total_class
    return np.mean(correct_list/total_list)


def mean_iou(predicted_img, labels, class_number):
    correct_list = np.zeros(class_number)
    FN_list = np.zeros(class_number)
    FP_list = np.zeros(class_number)
    for i in range(class_number):
        correct = np.logical_and(labels==predicted_img, labels==i).sum().item()
        correct_list[i] = correct

        FN_class = (predicted_img == i).sum().item()
        FN_list[i] = FN_class

        FP_class = (labels == i).sum().item()
        FP_list[i] = FP_class
    return np.mean(correct_list/(FN_list+FP_list-correct_list))


def cluster3D(image):
    time_start_cc = time.time()
    '''
    kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE,(2,2)) 
    opening=cv2.morphologyEx(image, cv2.MORPH_OPEN, kernel,iterations=1)
    fcco = '/home/haoyu/tomo_open.mrc'
    mrcfile.new(fcco, data=opening.astype('float32'), overwrite=True)
    '''
    print('==>start calculate conneted component ')

    labelimg,numcc=measure.label(image,connectivity=2,background =0,return_num=True)
    print('number of conneted components:',numcc)
    
    if os.path.exists('Lcc.txt'):
        os.remove('Lcc.txt')
    fc = open('Lcc.txt','a')

    kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE,(2,2))
    
    regions = measure.regionprops(labelimg)
    box = []
    
    sub_copy=image.copy()
    for prop in regions:
        #if prop.area<1*1*1:
        #            continue
        '''
        min_row, min_col,minz, max_row, max_col,maxz=prop.bbox
        if prop.area>50*50*50 or max_row-min_row>50 or max_col-min_col>50 or maxz-minz>50:
                    #print('area:',prop.area)
                    cordinate_large_area=np.where(labelimg==prop.label)
                    labelimg[cordinate_large_area]=0
                    a=cv2.morphologyEx(sub_copy, cv2.MORPH_OPEN, kernel,iterations=1)
                    labelimg[cordinate_large_area]=labelimg[cordinate_large_area]+a[cordinate_large_area]
        '''
        box.append(prop.centroid)
        strc=str(int(round(prop.centroid[2])))+' '+str(int(round(prop.centroid[1])))+' '+str(int(round(prop.centroid[0])))+'\n'   ##xyz
        fc.write(strc)
        fc.flush()
    fc.close
    labelimg[labelimg>0]=1
    image=image*labelimg



    print('===>start clustering')


    pimg = np.nonzero(image)
    X=np.array(list(zip(*pimg)))
    bandwidth=10 #8.2
    #bandwidth=8.2
    print('bandwidth',bandwidth)
    ms = MeanShift(bandwidth=bandwidth, seeds=box, n_jobs=-2, cluster_all=False,bin_seeding = True)
    #ms = MeanShift(bandwidth=bandwidth, seeds=box, n_jobs=-2, cluster_all=True)
    ms.fit(X)
    labels = ms.labels_
    cluster_centers = ms.cluster_centers_
    labels_unique = np.unique(labels)
    n_clusters_ = len(labels_unique)
    print("number of estimated clusters : %d" % n_clusters_)
     
    print('===>write locations')
    if os.path.exists('Location3D.txt'):
        os.remove('Location3D.txt')
    f = open('Location3D.txt','a')


    resultArr = np.zeros(np.shape(image), dtype=int)
    #print(np.shape(image),np.shape(resultArr))

    for k in range(n_clusters_-1):
        my_members = labels == k
        #print("cluster_size:",np.sum(my_members!=0))
        if np.sum(my_members!=0)<5*5*5:
                    continue
        cluster_center = cluster_centers[k]
        resultArr[X[my_members,0], X[my_members,1], X[my_members,2]]=k
        strin=str(int(round(cluster_center[2])))+' '+str(int(round(cluster_center[1])))+' '+str(int(round(cluster_center[0])))+'\n'   ##xyz
        f.write(strin)
        f.flush()
        #print(strin)
    f.close

    time_elapsed = time.time()-time_start_cc
    print('Time for clustering  {:.0f}m {:.0f}s'.format(time_elapsed // 60, time_elapsed % 60))
    fout = '/home/haoyu/tomo_cluster.mrc'
    mrcfile.new(fout, data=resultArr.astype('float32'), overwrite=True)

    return image


def runTest(image_dir, target_dir,  predict_dir,   trained_model_dir,  use_gpu=0):

    tomo_path='/home/haoyu/test_mrc/11_cluster.mrc'
    #tomo_path='/home/haoyu/data/SEG01/rec9.mrc'
    input_em=mrcfile.open(tomo_path, permissive=True);
    MRC=np.array(input_em.data)
    cluster3D(MRC)

def runMsdnet(image_dir, target_dir,  predict_dir,   trained_model_dir,  use_gpu=0):

    #input_tomo_path='/home/haoyu/data/MRC_norm_200/rec9.mrc'
    input_tomo_path='/home/haoyu/test_mrc/11_bin4.mrc'
    input_em=mrcfile.open(input_tomo_path, permissive=True);
    input_data=np.array(input_em.data)
    img_shape = np.shape(input_data)
    input_x,input_y,input_z=img_shape[2],img_shape[1],img_shape[0]
    print(input_x,input_y,input_z)

    diameter=25 #32
    half_dia=int(diameter/2)
    sld_itv=diameter*2-25
    print("clip_volume_size=",diameter*2)

    time_start = time.time()
    MRC=np.zeros([input_z, input_y, input_x],dtype=float)
    #print('===> Loading datasets')

    #test_set = DatasetFromFolder_mrc_test(image_dir, target_dir)
    #image_names = test_set.filenames
    #testing_data_loader = DataLoader(dataset=test_set, num_workers=8, batch_size=1, shuffle=False)

    print('===> loading model')

    Net = nn.DataParallel(model.Net(), device_ids=device_ids)

    #Net = model.Net()
    criterion = nn.MSELoss()
    if use_gpu:
        Net = Net.cuda()
    

    Net.load_state_dict(torch.load(trained_model_dir))
    print('===> inference and merging')    
    with torch.no_grad():
        i = 0

        for x_vol in range(diameter,input_x,sld_itv):
         if(x_vol+diameter>input_x):
            x_vol=input_x-diameter   
         for y_vol in range(diameter,input_y,sld_itv):
          if(y_vol+diameter>input_y):
              y_vol=input_y-diameter
          for z_vol in range(diameter,input_z,sld_itv):
            if(z_vol+diameter>input_z):
                #print(z_vol)
                z_vol=input_z-diameter
            print("xyz:",x_vol,y_vol,z_vol)
            input_mrc = input_data[z_vol-diameter:z_vol+diameter,y_vol-diameter:y_vol+diameter,x_vol-diameter:x_vol+diameter]
            input_shape=np.shape(input_mrc)
            '''
            path=str(x_vol)+'_'+str(y_vol)+'_'+str(z_vol)+'.mrc'
            fin=os.path.join('/home/haoyu/test_loc/',path)
            print(fin)
            mrcfile.new(fin, data=input_mrc.astype('float32'), overwrite=True)
            '''
            input_array= input_mrc.reshape((1,input_shape[0], input_shape[1], input_shape[2]))
            input=torch.from_numpy(input_array)
            norm=Compose([tio.RescaleIntensity(out_min_max=(0,1))])
            input=norm(input)
            
            input= input.reshape((1,1,input_shape[0], input_shape[1], input_shape[2]))
            
            prediction = Net(input)
            prediction = torch.argmax(prediction, dim = 1)
            
            prediction = prediction.type(torch.FloatTensor)
            prediction = prediction.view(input_shape[0], input_shape[1], input_shape[2])
            
            #merge
            dX,dY,dZ=input_shape[0], input_shape[1], input_shape[2]
            #tmpstr = image_names[i].split('_')
            sX,sY,sZ=x_vol,y_vol,z_vol#tmpstr[1],tmpstr[2],tmpstr[3].split('.')[0]
            #sX,sY,sZ=tmpstr[2],tmpstr[3],tmpstr[4].split('.')[0]
            startZ,endZ,startY,endY,startX,endX=int(int(sZ)-dZ/2),int(int(sZ)+dZ/2),int(int(sY)-dY/2),int(int(sY)+dY/2),int(int(sX)-dX/2),int(int(sX)+dX/2)
            #print(startZ,endZ,startY,endY,startX,endX)
            
            sub=prediction.numpy()
            sub_copy=sub.copy()
            #sub_image=np.zeros(np.shape(sub),dtype=float)

            #opening
            
            sublabel,subnum=measure.label(sub,connectivity=1,background =0,return_num=True)
            regions_sub = measure.regionprops(sublabel)
            
            #kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE,(2,2))
            for prop in regions_sub:
                #if prop.area<5*5*1:
                #    continue
                #print('area:',prop.area)
                '''
                min_row, min_col,minz, max_row, max_col,maxz=prop.bbox
                if prop.area>10*10*10 or max_row-min_row>10 or max_col-min_col>10 or maxz-minz>10:
                    #print('area:',prop.area)
                    cordinate_large_area=np.where(sublabel==prop.label)
                    sublabel[cordinate_large_area]=0
                    a=cv2.morphologyEx(sub_copy, cv2.MORPH_OPEN, kernel,iterations=1)
                    sublabel[cordinate_large_area]=sublabel[cordinate_large_area]+a[cordinate_large_area]
                '''
            sublabel[sublabel>0]=1
            sub=sub*sublabel
            

            #maskout except the tomo border
            MASK=np.copy(sub)
            if startX==0 :
               MASK[0:dZ,0:dY,0:1]=0
            if  endX==input_x :
               MASK[0:dZ,0:dY,dX-1:dX]=0
            if startY==0 :
               MASK[0:dZ,0:1,0:dX]=0
            if endY==input_y :
               MASK[0:dZ,dY-1:dY,0:dX]=0
            if startZ==0 :
               MASK[0:1,0:dY,0:dX]=0
            if endZ==input_z :
               MASK[dZ-1:dZ,0:dY,0:dX]=0
            segmentation.clear_border(MASK,in_place=True)
            sub=MASK*sub
            MRC[startZ:endZ,startY:endY,startX:endX]=  MRC[startZ:endZ,startY:endY,startX:endX] + sub
            i += 1
            if(z_vol==input_z-diameter):
               break
          if(y_vol==input_y-diameter):
              break
         if(x_vol==input_x-diameter):
            break
            #print('\n')
    MRC[MRC>0]=1
    fout ='/home/haoyu/test_mrc/tomo_cluster.mrc'
    print('==>finish merging.\n merged file in:',fout)
    mrcfile.new(fout, data=MRC.astype('float32'), overwrite=True)
    time_elapsed = time.time()-time_start
    print('Time for prediciton & merge {:.0f}m {:.0f}s'.format(time_elapsed // 60, time_elapsed % 60))
    #print('sizeMRC:',np.shape(MRC))
    #evaluation
    input_tomo_path='/home/haoyu/RealData/SEG/11_bin4.mrc'
    input_em=mrcfile.open(input_tomo_path, permissive=True);
    input_data=np.array(input_em.data)
    img_shape = np.shape(input_data)
    input_img = input_data.reshape(( img_shape[0], img_shape[1], img_shape[2]))
    #print('sizeinput:',np.shape(input_img))
    
    VA = global_accuracy(MRC, input_img)
    MVA = mean_accuracy(MRC, input_img, class_number)
    MIoU = mean_iou(MRC, input_img, class_number)
    
    
    print('VA=',VA)
    print('MVA=',MVA)
    print('MIoU=',MIoU)
    cluster3D(MRC)
    time_elapsed = time.time()-time_start
    print('Time elapse  {:.0f}m {:.0f}s'.format(time_elapsed // 60, time_elapsed % 60))
