import sys
import numpy as np
import matplotlib.pyplot as plt
import pickle
import os
import gzip

from torch.utils import data
import torch
import torch.optim as optim
from torch.autograd import Variable
from time import gmtime, strftime
import sys
import torch.nn as nn
from torch.utils.data import DataLoader

from data_utils import DatasetFromFolder_mrc_train, DatasetFromFolder_mrc_test
#from data_utils import DatasetSampler
from torch.utils.data.sampler import SubsetRandomSampler
from torch.cuda.amp import GradScaler

import model
#sys.path.append('..')
#from hicplusplus.hic_msdnet import msdmodel as msdmodel
#from Visualizer import Visualizer
#import msdmodel
import globalvar as gl
import config
import time
import os
class_number = gl.get_value('class_number')
num_layers = gl.get_value('num_layers')
kernel_size = gl.get_value('kernel_size')
batch_size = gl.get_value('batch_size')
learning_rate = gl.get_value('learning_rate')
growth_rate = int(gl.get_value('growth_rate'))
dilation_mod =  gl.get_value('dilation_mod')
log_dir = gl.get_value('log_dir')
modelsave_dir = gl.get_value('modelsave_dir')
epochs = gl.get_value('epochs')
device_ids = gl.get_value('device_ids')

def trainMsdnet( image_dir, target_dir, test_image_dir, test_target_dir,  pretrain_path = None, use_gpu=1):
    if use_gpu == 1:
        device = torch.device("cuda")
        torch.backends.cudnn.benchmark = True
        #torch.backends.cudnn.deterministic = True
    else:
        device = torch.device('cpu')

    print('===> Loading training datasets')

    train_set = DatasetFromFolder_mrc_train(image_dir, target_dir)
    training_data_loader = DataLoader(dataset=train_set, num_workers=24, batch_size=batch_size, shuffle=True, pin_memory=True, drop_last = True) #drop_last = True,

    print('===> Loading testing datasets')

    test_batch_size = 1
    #test_size = test_batch_num * batch_size  # test set size
    test_set = DatasetFromFolder_mrc_test(test_image_dir, test_target_dir)  # [0:test_size]
    testing_data_loader = DataLoader(dataset=test_set, num_workers=24, batch_size=test_batch_size, pin_memory=True) #, sampler=sampler_test)
    
    print('===> Building model')

    Net = nn.DataParallel(model.Net(), device_ids=device_ids)
    
    para = model.count_parameters(model.Net())
    print('Params num = {} = {:4f}K'.format( para , para * 4 / 1000) )
    print('Layers num = ', model.count_conv3d(model.Net()))
    print('Batch size = ',batch_size)
    print('max dilation = ',dilation_mod)
    print('learning_rate = ', learning_rate)
    print('gpu id = ', device_ids)
    print('gpu name = ',torch.cuda.get_device_name(0))

    criterion =  nn.NLLLoss() #nn.MSELoss() ## nn.CrossEntropyLoss()##
    if use_gpu:
        Net = Net.cuda()

    epoch_append = 0
    if pretrain_path != None:
        Net.load_state_dict(torch.load(pretrain_path))
        epoch_append = int(pretrain_path.split('_')[-1])
        
    optimizer = optim.Adam(Net.parameters(), lr = learning_rate)  # optimizer can be changed

    log = open(log_dir, 'a+')
    torch.cuda.synchronize()
    time_start = time.time()
 #   vis = Visualizer()
 #   vis = Visualizer(env='msd')
     
    scaler = GradScaler()
    print('===> start training')
    for epoch in range(1+epoch_append, epochs+1+epoch_append):
        epoch_loss = 0
        torch.cuda.empty_cache()
        for iteration, batch in enumerate(training_data_loader, 1):
            input = batch[0].to(device)
            #print('input min,max',torch.min(input), torch.max(input))
            target = batch[1]
            #print('target min,max',torch.min(target), torch.max(target))
            #print(target)
            target_size = target.size()
            target = target.type(torch.LongTensor)
            target = target.view(batch_size, target_size[-3], target_size[-2], target_size[-1])
            target = target.type(torch.LongTensor)
            target = target.to(device)

            optimizer.zero_grad()
            # Casts operations to mixed precision 
            #with torch.cuda.amp.autocast():
            net_output = Net(input)
            loss = criterion(net_output,  target)
            epoch_loss += loss.item()
            # Scales the loss, and calls backward() 
            # to create scaled gradients
            scaler.scale(loss).backward()
            scaler.step(optimizer)
            scaler.update()

            #loss.backward()
            #optimizer.step()
            #if (iteration % 20 == 0):
                #print("===> Epoch[{}]({}/{}):Training Batch Loss: {:.8f}".format(epoch, iteration, len(training_data_loader), loss.item()))
        print("===> Epoch {} Complete: Avg. Training Loss: {:.8f}".format(epoch, epoch_loss / len(training_data_loader))) 
        
        time_elapsed = time.time()-time_start
        print('Time elapse  {:.0f}m {:.0f}s'.format(time_elapsed // 60, time_elapsed % 60))
        print('===> start validation')
        torch.cuda.empty_cache()
        test_epoch_loss = 0    
        with torch.no_grad():
            for iteration, batch in enumerate(testing_data_loader, 1):
                input = batch[0].to(device)
                target = batch[1]
                #print('target max', torch.max(target))
                #input_var = torch.autograd.Variable(input)
                #target_var = torch.autograd.Variable(target)

                target_size = target.size()
                target = target.type(torch.LongTensor)
                target = target.view(test_batch_size, target_size[-3], target_size[-2], target_size[-1])
                target = target.type(torch.LongTensor)
                target = target.to(device)
                test_loss = criterion(Net(input), target)
                test_epoch_loss += test_loss.item()
                #print("===> Epoch[{}]({}/{}):Validation Batch Loss: {:.8f}".format(epoch, iteration, len(testing_data_loader),  test_loss.item()))
            print("===> Epoch {} Complete: Avg. Validation Loss: {:.8f}".format(epoch, test_epoch_loss / len(testing_data_loader)))
    
                
        loss_all = epoch_loss / len(training_data_loader)
        #vis.plot('Average Batch loss', loss_all)
        #record_log = {'epoch':epoch, 'Avg training batch loss':epoch_loss/len(training_data_loader)}
        #vis.log(record_log)
        torch.cuda.synchronize()
        time_elapsed = time.time()-time_start
        print('Training elapse  {:.0f}m {:.0f}s'.format(time_elapsed // 60, time_elapsed % 60))
        log.write(str(epoch) + ', ' + str(epoch_loss / len(training_data_loader)) + '\n')

        if (epoch % 1 == 0): # if (epoch % 10 == 0):
            torch.save(Net.state_dict(),  modelsave_dir+'pytorch_model_' + str(epoch))
            pass

        
    



